body {
  font-family: sans-serif;
  background: #fff;
  color: #000;
}

.note {
  margin-bottom: 2em;
}

img, video {
  max-height: 300px;
}

.metadata {
  border: none;
  margin: 1em 0;
}

.metadata td, .metadata th {
  border: none;
  text-align: left;
}

.metadata th, .metadata tr.direction td, .metadata tr.source td {
  font-weight: normal;
}

.metadata td, .metadata td span {
  font-weight: bold;
}
